{-# OPTIONS --copatterns --sized-types --guardedness #-}

module CCTree.Definitions where
open import Data.Product hiding (map)
open import Size public
open import Data.Unit
open import Data.Maybe hiding (_>>=_) renaming (map to mapMaybe)
open import Data.Bool
open import CTree using (CTree; ∞CTree; force; CTree' ; Concurrent ; _~̂[_]_; defaultPar; _⇄_ ; ⇄-sym ; ⇄-step) public
open import CTree.Transitions renaming (_[_]=>_ to _[_]=>'_ ; _[_]⇒_ to _[_]⇒'_) public
import CTree as CT
open import Relation.Binary.PropositionalEquality
open import Data.Nat
open import Relation.Nullary

infixl 7 _∥⃗_
infixl 7 _⊕_
infixl 5 _>>=_
infixl 5 _>>_

mutual
  data CCTree (E : Set → Set) (A : Set) (i : Size) : Set₁ where
    ret   : (v : A) → CCTree E A i
    later : (p : ∞CCTree E A i) → CCTree E A i
    _⊕_   : (p q : CCTree E A i) → CCTree E A i
    empty : CCTree E A i
    eff   : ∀ {B} → (e : E B) → (c : B → CCTree E A i) → CCTree E A i
    _>>=_ : ∀ {B} → CCTree E B i → (B → CCTree E A i) → CCTree E A i
    _∥⃗_ : ∀ {B} → CCTree E B i → CCTree E A i → CCTree E A i
    interpSt : ∀ {F S}  {{_ : Concurrent F}} → S → (∀ {B} → S → F B → CTree E (B × S) ∞) → CCTree F A i → CCTree E A i

  record ∞CCTree E (A : Set) (i : Size) : Set₁ where
    coinductive
    constructor delay
    field
      force : {j : Size< i} → CCTree E A j

open ∞CCTree public

_∞>>=_ :  ∀ {i A B E} → ∞CCTree E A i → (A → CCTree E B i) → ∞CCTree E B i
force (x ∞>>= f) = force x >>= f


-- Monadic bind operator.

mutual 
  ⟦_⟧ : ∀ {i A E} {{_ : Concurrent E}} → CCTree E A i → ∀ {R} → (A → CTree E R i) → CTree E R i
  ⟦ ret v ⟧ k = k v
  ⟦ later p ⟧ k = CT.later (⟦ p ⟧∞ k)
  ⟦ p ⊕ q ⟧ k = ⟦ p ⟧ k CT.⊕ ⟦ q ⟧ k
  ⟦ empty ⟧ k = CT.empty
  ⟦ eff e c ⟧ k = CT.eff e (λ r → ⟦ c r ⟧ k)
  ⟦ p >>= f ⟧ k = ⟦ p ⟧ (λ r → ⟦ f r ⟧ k)
  ⟦ p ∥⃗ q ⟧ k =  ⟦ p ⟧ CT.ret CT.∥⃗ ⟦ q ⟧ k
  ⟦ interpSt s f p ⟧ k = CT.interpSt s f (⟦ p ⟧ CT.ret) CT.>>= k
  
  ⟦_⟧∞ : ∀ {i A E} {{_ : Concurrent E}} → ∞CCTree E A i → ∀ {R} → (A → CTree E R i) → ∞CTree E R i
  force (⟦ p ⟧∞ k) = ⟦ force p ⟧ k

_>>_ : ∀ {i A B E} → CCTree E A i → CCTree E B i → CCTree E B i
p >> q = p >>= λ _ → q

-- Monadic return operator

return : ∀ {i A E} → A → CCTree E A i
return = ret

-- Functorial map operator derived from >>=

map : ∀ {i A B E} → (A → B) → CCTree E A i → CCTree E B i
map f p = p >>= (λ x → return (f x))


mutual
  never : ∀ {i A E} -> CCTree E A i
  never = later ∞never

  ∞never : ∀ {i A E} -> ∞CCTree E A i
  force ∞never = never

------------
-- interp --
------------

-- This is the standard effect interpretation function (without
-- state).

interp : ∀ {i A E} {{_ : Concurrent E}} → (∀ {B} → E B → CTree E B ∞) → CCTree E A i → CCTree E A i
interp f = interpSt tt (λ s x → CT.map (λ y → y , tt) (f x))
