{-# OPTIONS --copatterns --sized-types --guardedness #-}

module CCTree.IndexedBisimilarity where
open import CCTree.Definitions public
import CTree.Bisimilarity as CT
open import Data.Product hiding (map)
open import Size public
open import Data.Unit
import CTree as CT
open import Data.Nat
open import Data.Maybe hiding (_>>=_) renaming (map to mapMaybe)
open import Relation.Binary.PropositionalEquality

---------------------------------------------
-- Definition of step-indexed bisimilarity --
---------------------------------------------

variable
  i : ℕ
  A : Set
  B : Set
  A' : Set
  B' : Set
  C : Set
  S : Set
  E : Set → Set
  F : Set → Set



infix 3 _~[_]_
record _~[_]_ {{_ : Concurrent E}} (p : CCTree E A ∞) (i : ℕ) (q : CCTree E A ∞) : Set₁ where
  constructor ~imk
  field
    ~iapp :  ∀ {R} (k : A → CTree E R ∞) → ⟦ p ⟧ k CT.~[ i ] ⟦ q ⟧ k

open _~[_]_ public

--------------------------------
-- Definition of bisimilarity --
--------------------------------

infix 3 _~_
record _~_  {{_ : Concurrent E}} (p : CCTree E A ∞) (q : CCTree E A ∞) : Set₁ where
  constructor ~mk
  field
    ~app :  ∀ {R} (k : A → CTree E R ∞) → ⟦ p ⟧ k CT.~ ⟦ q ⟧ k

open _~_ public

~-~i : ∀ {{_ : Concurrent E}} {p q : CCTree E A ∞} → (p ~ q) → p ~[ i ] q
~-~i (~mk eq) = ~imk λ k → CT.~-~i (eq k)

~i-~ : ∀ {{_ : Concurrent E}} {p q : CCTree E A ∞} → (∀ i → p ~[ i ] q) → p ~ q
~i-~ eqi = ~mk λ k → CT.~i-~ λ i → ~iapp (eqi i) k

-----------------------
-- basice properties --
-----------------------

~izero : ∀ {{_ : Concurrent E}} {p q : CCTree E A ∞} → p ~[ 0 ] q
~izero = ~imk λ k → CT.~izero

~irefl : ∀ {{_ : Concurrent E}} {p : CCTree E A ∞} → p ~[ i ] p
~irefl = ~imk (λ k → CT.~irefl)

~isym : ∀ {{_ : Concurrent E}} {p q : CCTree E A ∞} → p ~[ i ] q → q ~[ i ] p
~isym (~imk b) = ~imk λ k → CT.~isym (b k)

~itrans : ∀ {{_ : Concurrent E}} {p q r : CCTree E A ∞} → p ~[ i ] q → q ~[ i ] r → p ~[ i ] r
~itrans (~imk b1) (~imk b2) = ~imk λ k → CT.~itrans (b1 k) (b2 k)


------------------------
-- calculation syntax --
------------------------

_~⟨_⟩_ : ∀ {{_ : Concurrent E}} (x : CCTree E A ∞) →
  ∀ {y : CCTree E A ∞} {z : CCTree E A ∞} → x ~[ i ] y → y ~[ i ] z → x ~[ i ] z
_~⟨_⟩_ {_} x r eq =  ~itrans r eq

_~⟨⟩_ : ∀ {{_ : Concurrent E}} (x : CCTree E A ∞) → ∀ {y : CCTree E A ∞} → x ~[ i ] y → x ~[ i ] y
_~⟨⟩_  x eq = eq


_∎ : ∀ {{_ : Concurrent E}} (x : CCTree E A ∞) → x ~[ i ] x
_∎  x = ~irefl

infix  3 _∎
infixr 1 _~⟨_⟩_
infixr 1 _~⟨⟩_


-----------------
-- congruences --
-----------------

~icong : ∀ {{_ : Concurrent E}} (p : CCTree E A ∞) {k k' : A → CTree E B ∞}
  (b : ∀ x → k x CT.~[ i ] k' x) → ⟦ p ⟧ k CT.~[ i ] ⟦ p ⟧ k'
~icong {i = 0} _ _ = CT.~izero
~icong (ret v) b = b v
~icong {i = suc i} (later p) b = CT.~ilater (~icong {i = i} (force p) (λ x → CT.~isuc (b x)))
~icong (p ⊕ q) b = CT.⊕-cong (~icong p b) (~icong q b)
~icong empty b = CT.~irefl
~icong (eff e c) b = CT.~ieff e (λ r → ~icong (c r) b)
~icong (p >>= f) b = ~icong p λ r → ~icong (f r) b
~icong (p ∥⃗ q) b = CT.∥⃗-cong-r (~icong q b)
~icong (interpSt s f p) b = CT.>>=-cong-r (CT.interpSt s f (⟦ p ⟧ CT.ret)) b


~icong-map : ∀ {{_ : Concurrent E}} (p : CCTree E A ∞) {f : B → C} → {k : A → CTree E B ∞} →
  CT.map f (⟦ p ⟧ k)  CT.~[ i ] ⟦ p ⟧ (λ r → CT.map f (k r))
~icong-map {i = 0} _ = CT.~izero
~icong-map (ret v) = CT.~irefl
~icong-map {i = suc i} (later p) = CT.~ilater (~icong-map {i = i} (force p))
~icong-map (p ⊕ q) = CT.⊕-cong (~icong-map p) (~icong-map q)
~icong-map empty = CT.~irefl
~icong-map (eff e c) = CT.~ieff e (λ r → ~icong-map (c r))
~icong-map (p >>= g) = CT.~itrans (~icong-map p) (~icong p (λ r → ~icong-map (g r))) 
~icong-map (p ∥⃗ q) = CT.~itrans (CT.∥⃗-map-r (⟦ p ⟧ CT.ret) _) (CT.∥⃗-cong-r (~icong-map q))
~icong-map (interpSt s h p) = CT.>>=-assoc (CT.interpSt s h (⟦ p ⟧ CT.ret))


>>=-cong : ∀ {{_ : Concurrent E}}  {p q : CCTree E A ∞} (b : p ~[ i ] q)
            {k l : A → CCTree E B ∞} →
            (h : ∀ a → k a ~[ i ] l a) →
            (p >>= k) ~[ i ] (q >>= l)
>>=-cong {q = q} b {k} {l} h = ~imk (λ k'  → CT.~itrans (~iapp b (λ r → ⟦ k r ⟧ k')) (~icong q λ x → ~iapp (h x) k'))

>>=-cong-r : ∀ {{_ : Concurrent E}} (a : CCTree E A ∞)
          {k l : A → CCTree E B ∞} (h : ∀ a → k a ~[ i ] l a) →
          (a >>= k) ~[ i ] (a >>= l)
>>=-cong-r a h = >>=-cong ~irefl h

>>-cong-r : ∀ {{_ : Concurrent E}} (a : CCTree E A ∞)
          {k l : CCTree E B ∞} (h : k ~[ i ] l) →
          (a >> k) ~[ i ] (a >> l)
>>-cong-r a h = >>=-cong-r a (λ _ → h)


>>=-cong-l : ∀ {{_ : Concurrent E}} {p q : CCTree E A ∞} (b : p ~[ i ] q)
             {k : A → CCTree E B ∞} →
             (p >>= k) ~[ i ] (q >>= k)
>>=-cong-l b = >>=-cong b (λ _ → ~irefl)

map-cong : ∀ {{_ : Concurrent E}}  {p q : CCTree E A ∞} (b : p ~[ i ] q)
                {f : A → B} → map f p ~[ i ] map f q
map-cong b = >>=-cong-l b


~ilater : ∀ {{_ : Concurrent E}} {a b : ∞CCTree E A ∞} → force a ~[ i ] force b → later a ~[ suc i ] later b
~ilater (~imk b) = ~imk λ k → CT.~ilater (b k)


⊕-cong : ∀ {{_ : Concurrent E}} {p1 q1 p2 q2 : CCTree E A ∞} → p1 ~[ i ] p2 → q1 ~[ i ] q2
  → p1 ⊕ q1 ~[ i ] p2 ⊕ q2
⊕-cong (~imk b1) (~imk b2) = ~imk λ k → CT.⊕-cong (b1 k) (b2 k)

⊕-cong-r : ∀ {{_ : Concurrent E}} {p q q' : CCTree E A ∞} → q ~[ i ] q' → p ⊕ q ~[ i ] p ⊕ q'
⊕-cong-r ~q = ⊕-cong ~irefl ~q


⊕-cong-l : ∀ {{_ : Concurrent E}} {p q p' : CCTree E A ∞} → p ~[ i ] p' → p ⊕ q ~[ i ] p' ⊕ q
⊕-cong-l ~p =  ⊕-cong ~p ~irefl


--------------------------
-- properties about >>= --
--------------------------

never->>= : ∀ {{_ : Concurrent E}} {f : A → CCTree E B ∞} → (never >>= f) ~[ i ] never
never->>= {i = 0} = ~izero
never->>= {i = suc i} {f = f} = ~imk λ k → CT.~ilater (~iapp (never->>= {f = f}) k)

>>=-later : ∀ {{_ : Concurrent E}} {p : ∞CCTree E A ∞} {f : A → CCTree E B ∞}
  → (later p >>= f) ~[ i ] later (p ∞>>= f)
>>=-later {i = zero} = ~izero
>>=-later {i = suc i} {p = p} {f} = ~imk λ k → CT.~ilater CT.~irefl


----------------
-- monad laws --
----------------

>>=-return : ∀ {{_ : Concurrent E}} {p : CCTree E A ∞}  → (p >>= return) ~[ i ] p
>>=-return = ~imk (λ k → CT.~irefl)

return->>= : ∀ {{_ : Concurrent E}} {i} {x : A} {f : A → CCTree E B ∞}  → (return x >>= f) ~[ i ] f x
return->>= = ~imk (λ k → CT.~irefl)

>>=-assoc : ∀ {{_ : Concurrent E}} (p : CCTree E A ∞)
                {k : A → CCTree E B ∞}{l : B → CCTree E C ∞} →
                ((p >>= k) >>= l) ~[ i ] (p >>= λ a → k a >>= l)
>>=-assoc p {k} {l} = ~imk λ k → CT.~irefl


>>=-assoc' : ∀ {{_ : Concurrent E}} (p : CCTree E A ∞)
                {k : A → CCTree E B ∞}{l : B → CCTree E C ∞}{f : A → CCTree E C ∞} →
                (∀ a → k a >>= l ~[ i ] f a) →
                ((p >>= k) >>= l) ~[ i ] (p >>= f)
>>=-assoc' p b = ~itrans (>>=-assoc p) (>>=-cong-r p b)

>>-assoc' : ∀ {{_ : Concurrent E}} (p : CCTree E A ∞)
                {k : CCTree E B ∞}{l : B → CCTree E C ∞}{f : CCTree E C ∞} →
                (k >>= l ~[ i ] f) →
                ((p >> k) >>= l) ~[ i ] (p >> f)
>>-assoc' p b = ~itrans (>>=-assoc p) (>>-cong-r p b)


>>-assoc : ∀ {{_ : Concurrent E}} (p : CCTree E A ∞)
             {q : CCTree E B ∞}{r : CCTree E C ∞} →
             (p >> q) >> r ~[ i ] p >> (q >> r)
>>-assoc p = >>=-assoc p

map->>= : ∀ {{_ : Concurrent E}} (p : CCTree E A ∞)
                {k : A → B}{l : B → CCTree E C ∞} →
                ((map k p) >>= l) ~[ i ] (p >>= λ a → l (k a))
map->>= p {k} {l} = ~itrans (>>=-assoc p) (>>=-cong-r p (λ r → return->>=))


--------------------------
-- non-determinism laws --
--------------------------


⊕-unit-l : ∀ {{_ : Concurrent E}} {p : CCTree E A ∞} → empty ⊕ p ~[ i ] p
⊕-unit-l = ~imk λ k → CT.⊕-unit-l
 
⊕-assoc : ∀ {{_ : Concurrent E}} {p q r : CCTree E A ∞} → (p ⊕ q) ⊕ r ~[ i ] p ⊕ (q ⊕ r)
⊕-assoc = ~imk λ k → CT.⊕-assoc

⊕-idem : ∀ {{_ : Concurrent E}} {p : CCTree E A ∞} → p ⊕ p ~[ i ] p
⊕-idem = ~imk λ k → CT.⊕-idem

⊕-comm : ∀ {{_ : Concurrent E}} {p q : CCTree E A ∞} → p ⊕ q ~[ i ] q ⊕ p
⊕-comm = ~imk λ k → CT.⊕-comm

⊕-unit-r : ∀ {{_ : Concurrent E}} {p : CCTree E A ∞} → p ⊕ empty ~[ i ] p
⊕-unit-r = ~imk λ k → CT.⊕-unit-r

⊕-distr : ∀ {{_ : Concurrent E}} (p q : CCTree E A ∞) {f : A → CCTree E B ∞}
  → (p ⊕ q) >>= f ~[ i ] (p >>= f) ⊕ (q >>= f)
⊕-distr p q = ~imk λ k → CT.~irefl

empty->>= :  ∀ {{_ : Concurrent E}} {f : A → CCTree E B ∞} → empty >>= f ~[ i ] empty
empty->>= = ~imk λ k → CT.~irefl

-------------------
-- parallel laws --
-------------------

return-∥⃗ : ∀ {{_ : Concurrent E}} {v : A} {p : CCTree E B ∞} 
  → return v ∥⃗ p ~[ i ] p
return-∥⃗ = ~imk λ k → CT.return-∥⃗


∥⃗->>= : ∀ {{_ : Concurrent E}} {p : CCTree E A ∞} {q : CCTree E B ∞} {f : B → CCTree E C ∞}
  → (p ∥⃗ q) >>= f ~[ i ] p ∥⃗ (q >>= f)
∥⃗->>= {q = q} {f} = ~imk λ k → CT.~irefl


∥⃗-cong : ∀ {{_ : Concurrent E}} {p p' : CCTree E A ∞}{q q' : CCTree E B ∞}
  → p ~[ i ] p' → q ~[ i ] q' → p ∥⃗ q ~[ i ] p' ∥⃗ q'
∥⃗-cong (~imk b1) (~imk b2) = ~imk λ k → CT.∥⃗-cong (b1 (CT.ret)) (b2 k)

∥⃗-cong-l : ∀ {{_ : Concurrent E}} {p p' : CCTree E A ∞}{q : CCTree E B ∞}
  → p ~[ i ] p' → p ∥⃗ q ~[ i ] p' ∥⃗  q
∥⃗-cong-l b = ∥⃗-cong b ~irefl

∥⃗-cong-r : ∀ {{_ : Concurrent E}} {p : CCTree E A ∞}{q q' : CCTree E B ∞}
  → q ~[ i ] q' → p ∥⃗ q ~[ i ] p ∥⃗ q'
∥⃗-cong-r b = ∥⃗-cong ~irefl b

~icong' : ∀ {{_ : Concurrent E}} (p : CCTree E A ∞) {k : A → CTree E B ∞} {k' : A → CTree E C ∞}
  (b : ∀ x → k x CT.>> CT.ret tt CT.~[ i ] k' x CT.>> CT.ret tt) → ⟦ p ⟧ k CT.>> CT.ret tt CT.~[ i ] ⟦ p ⟧ k' CT.>> CT.ret tt
~icong' p b = CT.~itrans (~icong-map p {f = λ _ → tt}) (CT.~itrans (~icong p b) (CT.~isym (~icong-map p {f = λ _ → tt})))


~icont : ∀ {{_ : Concurrent E}} (p : CCTree E A ∞) (f : A → B) →
         ⟦ p ⟧ CT.ret  CT.>> CT.ret tt CT.~[ i ] (⟦ p ⟧ (λ r → CT.ret (f r))) CT.>> CT.ret tt
~icont {i = 0} _ _ = CT.~izero
~icont (ret v) f = CT.~irefl
~icont {i = suc i}(later p) f = CT.~ilater (~icont {i = i}(force p) f)
~icont (p ⊕ q) f = CT.⊕-cong (~icont p f) (~icont q f)
~icont empty f = CT.~irefl
~icont (eff e c) f = CT.~ieff e (λ r → ~icont (c r) f)
~icont (p >>= k) f = ~icong' p (λ x → ~icont (k x) f)
~icont (p ∥⃗ q) f = CT.~itrans (CT.∥⃗-map-r (⟦ p ⟧ CT.ret) (⟦ q ⟧ CT.ret) )
  (CT.~itrans (CT.∥⃗-cong-r (~icont q f)) (CT.~isym ((CT.∥⃗-map-r (⟦ p ⟧ CT.ret) (⟦ q ⟧ (λ r → CT.ret (f r)))))))
~icont (interpSt s h p) f = CT.~itrans (CT.>>=-assoc (CT.interpSt s h (⟦ p ⟧ CT.ret)))
  (CT.~isym (CT.>>=-assoc (CT.interpSt s h (⟦ p ⟧ CT.ret))))

∥⃗-map-l : ∀ {{_ : Concurrent E}} (p : CCTree E A ∞) (q : CCTree E B ∞) {f : A → C}
  → p ∥⃗ q ~[ i ]  map f p ∥⃗ q
∥⃗-map-l p q {f} = ~imk λ k → CT.~itrans (CT.~itrans (CT.∥⃗-map-l (⟦ p ⟧ CT.ret) (⟦ q ⟧ k) {f = λ x → tt})
  (CT.∥⃗-cong-l  (~icont p f ))) (CT.~isym ( CT.∥⃗-map-l (⟦ p ⟧ (λ r → CT.ret (f r))) (⟦ q ⟧ k) {f = λ x → tt}))  

∥⃗-map : ∀ {{_ : Concurrent E}} (p : CCTree E A ∞) (q : CCTree E B ∞) {f : A → A'} {g : B → B'}
  → map g (p ∥⃗ q) ~[ i ]  map f p ∥⃗ map g q
∥⃗-map p q  = ~itrans (∥⃗->>= {p = p} {q = q}) (∥⃗-map-l p _)

∥⃗-assoc : ∀ {{_ : Concurrent E}} (p : CCTree E A ∞) (q : CCTree E B ∞) (r : CCTree E C ∞)
  → (p ∥⃗ q) ∥⃗ r ~[ i ] p ∥⃗ (q ∥⃗ r)
∥⃗-assoc p q r = ~imk λ k → CT.∥⃗-assoc (⟦ p ⟧ CT.ret) (⟦ q ⟧ CT.ret) (⟦ r ⟧ k)

∥⃗-comm : ∀ {{_ : Concurrent E}} (p : CCTree E A ∞) (q : CCTree E B ∞) (r : CCTree E C ∞)
  → (p ∥⃗ q) ∥⃗ r ~[ i ] (q ∥⃗ p) ∥⃗ r
∥⃗-comm p q r = ~imk λ k → CT.∥⃗-comm (⟦ p ⟧ CT.ret) (⟦ q ⟧ CT.ret) (⟦ r ⟧ k)


------------
-- interp --
------------


interpSt-cong : ∀ {{_ : Concurrent E}} {{_ : Concurrent F}} {p q : CCTree E A ∞} {st : S} {f : ∀ {B} → S → E B → CTree F (B × S) ∞}
  → p ~[ i ] q → interpSt st f p ~[ i ] interpSt st f q
interpSt-cong (~imk b) = ~imk λ k → CT.>>=-cong-l (CT.interpSt-cong (b CT.ret))

interpSt-map : ∀ {{_ : Concurrent E}} {{_ : Concurrent F}} {p : CCTree E A ∞} {st : S} {f : ∀ {B} → S → E B → CTree F (B × S) ∞}
  (g : A → B) → map g (interpSt st f p) ~[ i ] interpSt st f (map g p)
interpSt-map {p = p}{st}{f} g = ~imk λ k → CT.~itrans
  (CT.~isym ((CT.map->>= (CT.interpSt st f (⟦ p ⟧ CT.ret)))  {k = g} {l = k}))
  (CT.>>=-cong-l {p = CT.map g (CT.interpSt st f (⟦ p ⟧ CT.ret))↑} {q = CT.interpSt st f (⟦ p ⟧ (λ r → CT.ret (g r)))↑}
  ( CT.~itrans (CT.interpSt-map (⟦ p ⟧ CT.ret) g) ((CT.interpSt-cong (~icong-map p)))) {k = k})
